import 'dart:convert';
import 'dart:typed_data';
import 'package:convert/convert.dart';

import 'transaction.dart';
import '../identities/address.dart';
import '../enums/types.dart';
import '../utils/buffer.dart';

Transaction deserialize(String serialized) {
  Uint8List bytes = hex.decode(serialized);
  BufferReader reader = new BufferReader(bytes);
  Transaction transaction = new Transaction();

  _deserializeHeader(transaction, reader);
  _deserializeType(transaction, reader);
  _deserializeSignatures(transaction, reader);

  if (transaction.secondSignature != null) {
    transaction.signSignature = transaction.secondSignature;
  }

  if (transaction.type == Types.vote || transaction.type == Types.secondSignature) {
    transaction.recipientId = Address.fromPublicKey(transaction.senderPublicKey, transaction.network).toString();
  }

  if (transaction.vendorFieldHex != null) {
    transaction.vendorField = utf8.decode(hex.decode(transaction.vendorFieldHex));
  }

  if (transaction.id == null) {
    transaction.id = transaction.getId();
  }

  return transaction;
}

void _deserializeHeader(Transaction transaction, BufferReader reader) {
  transaction.header = reader.getInt();
  transaction.version = reader.getInt();
  transaction.network = reader.getInt();
  transaction.type = reader.getInt();
  transaction.timestamp = reader.getInt32();
  transaction.senderPublicKey = reader.getHex(33);
  transaction.fee = reader.getInt64();

  int vendorFieldLength = reader.getInt();
  if (vendorFieldLength > 0) {
    transaction.vendorFieldHex = reader.getHex(vendorFieldLength);
  }
}

void _deserializeType(Transaction transaction, BufferReader reader) {
  switch (transaction.type) {
    case Types.delegateRegistration:
      return _deserializeDelegateRegistration(transaction, reader);
    case Types.secondSignature:
      return _deserializeSecondSignatureRegistration(transaction, reader);
    case Types.transfer:
      return _deserializeTransfer(transaction, reader);
    case Types.vote:
      return _deserializeVote(transaction, reader);
  }
}

void _deserializeDelegateRegistration(Transaction transaction, BufferReader reader) {
  int usernameLength = reader.getInt();
  Uint8List username = reader.get(usernameLength);

  transaction.asset = {
    'delegate': {
      'username': utf8.decode(username)
    }
  };
}

void _deserializeSecondSignatureRegistration(Transaction transaction, BufferReader reader) {
  String publicKey = reader.getHex(33);
  transaction.asset = {
    'signature': {
      'publicKey': publicKey
    }
  };
}

void _deserializeTransfer(Transaction transaction, BufferReader reader) {
  transaction.amount = reader.getInt64();
  transaction.expiration = reader.getInt32();
  Uint8List addressBytes = reader.get(21);
  transaction.recipientId = new Address(addressBytes).toString();
}

void _deserializeVote(Transaction transaction, BufferReader reader) {
  int votesLength = reader.getInt();
  List votes = new List<String>();

  for (int i = 0; i < votesLength; i++) {
    int voteType = reader.getInt();
    String voteTypeSymbol = voteType == 1 ? '+' : '-';

    StringBuffer vote = new StringBuffer();
    vote.write(voteTypeSymbol);
    vote.write(reader.getHex(33));

    votes.add(vote.toString());
  }

  transaction.asset = {
    'votes': votes
  };
}

void _deserializeSignatures(Transaction transaction, BufferReader reader) {
  String raw = reader.getHex();

  if (raw == null || raw.isEmpty) {
    return;
  }

  int signatureLength = int.parse(raw.substring(2, 4), radix: 16) + 2;
  signatureLength *= 2;
  transaction.signature = raw.substring(0, signatureLength);

  if (raw.length > signatureLength) {
    String secondSignature = raw.substring(signatureLength);
    transaction.secondSignature = secondSignature;
  }
}