import 'dart:convert';
import 'dart:typed_data';

import '../identities/address.dart';
import '../enums/types.dart';
import '../configuration/network.dart';
import '../utils/buffer.dart';
import 'transaction.dart';

Uint8List serialize(Transaction tx) {
  Buffer buffer = new Buffer();
  
  buffer.putInt(0xff);
  
  int version = tx.version != null && tx.version > 0 ? tx.version : 0x01;
  buffer.putInt(version);

  int network = tx.network != null && tx.network > 0 ? tx.network : NetworkConfig.get().version();
  buffer.putInt(network);

  buffer.putInt(tx.type);
  buffer.putInt32(tx.timestamp);
  buffer.putHex(tx.senderPublicKey);
  buffer.putInt64(tx.fee);

  _serializeVendorField(tx, buffer);
  _serializeType(tx, buffer);
  _serializeSignatures(tx, buffer);

  return buffer.toBytes();
}

void _serializeVendorField(Transaction tx, Buffer buf) {
  if (tx.vendorField != null) {
    buf.putInt(tx.vendorField.length);
    buf.put(utf8.encode(tx.vendorField));
  } else if (tx.vendorFieldHex != null) {
    buf.putInt(tx.vendorFieldHex.length ~/ 2);
    buf.putHex(tx.vendorFieldHex);
  } else {
    buf.putInt(0x00);
  }
}

void _serializeType(Transaction tx, Buffer buf) {
  switch (tx.type) {
    case Types.delegateRegistration:
      return _serializeDelegateRegistration(tx, buf);
    case Types.secondSignature:
      return _secondSignatureRegistration(tx, buf);
    case Types.transfer:
      return _serializeTransfer(tx, buf);
    case Types.vote:
      return _serializeVote(tx, buf);
  }
}

void _serializeDelegateRegistration(Transaction tx, Buffer buf) {
  String username = tx.asset['delegate']['username'];
  buf.putInt(username.length);
  buf.put(utf8.encode(username));
}

void _secondSignatureRegistration(Transaction tx, Buffer buf) {
  buf.putHex(tx.asset['signature']['publicKey']);
}

void _serializeTransfer(Transaction tx, Buffer buf) {
  buf.putInt64(tx.amount);
  buf.putInt32(tx.expiration);
  buf.put(Address.decode(tx.recipientId));
}

void _serializeVote(Transaction tx, Buffer buf) {
  List<String> votes = tx.asset['votes'];

  for (int i = 0;i < votes.length;i++) {
    votes[i] = (votes[i].startsWith('+') ? '01' : '00') + votes[i].substring(1);
  }

  buf.putInt(votes.length);
  buf.putHex(votes.join(''));
}

void _serializeSignatures(Transaction tx, Buffer buf) {
  if (tx.signature != null) {
    buf.putHex(tx.signature);
  }

  if (tx.secondSignature != null) {
    buf.putHex(tx.secondSignature);
  } else if (tx.signSignature != null) {
    buf.putHex(tx.signSignature);
  }

  if (tx.signatures != null) {
    buf.putInt(0xff);
    buf.putHex(tx.signatures.join(''));
  }
}