
import 'dart:typed_data';
import 'dart:convert';
import 'package:crypto/crypto.dart';
import 'package:convert/convert.dart';
import 'package:bs58check/bs58check.dart' as bs58check;
import 'package:bip32/src/utils/ecurve.dart' as ecc;

import '../enums/types.dart';
import '../utils/buffer.dart';
import '../identities/private_key.dart';
import '../identities/public_key.dart';
import '../utils/signature.dart';

class Transaction {
  // Header
  int header;
  int version;
  int network;
  int type;
  int timestamp;

  // Types
  String id;
  String senderPublicKey;
  String recipientId;
  String vendorField;
  String vendorFieldHex;
  int amount = 0;
  int fee = 0;

  // Signatures
  String signature;
  String secondSignature;
  String signSignature;
  List<String> signatures;

  int expiration;
  Map<String, dynamic> asset;

  Transaction();

  Transaction sign(String passphrase) {
    PrivateKey pk = PrivateKey.fromPassphrase(passphrase);
    PublicKey pb = new PublicKey(pk.getPublicKey());
    this.senderPublicKey = pb.toHex();
    Uint8List hash = this.toHash(true, true);
    Signature signature = Signature.decode(ecc.sign(hash, pk.toBytes()));
    this.signature = hex.encode(signature.toDER());
    return this;
  }

  Transaction secondSign(String passphrase) {
    PrivateKey pk = PrivateKey.fromPassphrase(passphrase);
    Uint8List hash = this.toHash(false, true);
    Signature signature = Signature.decode(ecc.sign(hash, pk.toBytes()));
    this.signSignature = hex.encode(signature.toDER());
    return this;
  }

  bool verify() {
    Uint8List hash = this.toHash(true, true);
    Uint8List publicKey = hex.decode(this.senderPublicKey);
    Signature signature = Signature.fromDER(hex.decode(this.signature));

    return ecc.verify(hash, publicKey, signature.encode());
  }

  bool secondVerify(String secondPublicKey) {
    Uint8List hash = this.toHash(true);
    Uint8List publicKey = hex.decode(secondPublicKey);
    Signature signature = Signature.fromDER(hex.decode(this.signature));

    return ecc.verify(hash, publicKey, signature.encode());
  }

  String getId() {
    return hex.encode(this.toHash());
  }

  Uint8List toHash([bool skipSignaure = false, bool skipSecondSignature = false]) {
    return sha256.convert(this.toBytes(skipSignaure, skipSecondSignature)).bytes;
  }
  
  Uint8List toBytes([bool skipSignaure = false, bool skipSecondSignature = false]) {
    Buffer buffer = new Buffer();
    buffer.putInt(this.type);
    buffer.putInt32(this.timestamp);
    buffer.putHex(this.senderPublicKey);

    bool skipRecipientId = this.type == Types.secondSignature || this.type == Types.multiSignature;

    if (this.recipientId != null && this.recipientId.isNotEmpty && !skipRecipientId) {
      buffer.put(bs58check.decode(this.recipientId));
    } else {
      buffer.put(new Uint8List(21));
    }

    Uint8List vendorBytes = new Uint8List(0);
    if (this.vendorFieldHex != null && this.vendorFieldHex.isNotEmpty) {
      vendorBytes = hex.decode(this.vendorFieldHex);
    } else if (this.vendorField != null && this.vendorField.isNotEmpty) {
      vendorBytes = utf8.encode(this.vendorField);
    }

    if (vendorBytes.length > 0 && vendorBytes.length <= 255) {
      buffer.put(vendorBytes);
      if (vendorBytes.length < 64) {
        buffer.put(new Uint8List(64-vendorBytes.length));
      }
    } else {
      buffer.put(new Uint8List(64));
    }

    buffer.putInt64(this.amount);
    buffer.putInt64(this.fee);

    if (this.asset != null && this.asset.isNotEmpty) {
      if (this.type == Types.secondSignature) {
        buffer.putHex(this.asset['signature']['publicKey']);
      } else if (this.type == Types.delegateRegistration) {
        buffer.putString(this.asset['delegate']['username']);
      } else if (this.type == Types.vote) {
        buffer.putString((this.asset['votes'] as List<String>).join(''));
      }
    }

    if (!skipSignaure && signature != null) {
      buffer.putHex(this.signature);
    }

    if (!skipSecondSignature && this.signSignature != null) {
      buffer.putHex(this.signSignature);
    }

    return buffer.toBytes();
  }

  Map<String, dynamic> toMap() {
    Map map = {
      'network': this.network,
      'id': this.id,
      'timestamp': this.timestamp,
      'expiration': this.expiration,
      'type': this.type,
      'amount': this.amount,
      'fee': this.fee,
      'recipientId': this.recipientId,
      'signature': this.signature,
      'senderPublicKey': this.senderPublicKey,
      'signSignature': this.signSignature,
      'vendorField': this.vendorField,
      'asset': this.asset
    };

    map.removeWhere((key, value) => value == null);

    return map;
  }

  String toJson() {
    return json.encode(this.toMap());
  }
}