import 'dart:typed_data';
import 'buffer.dart';

// NOTE: adapted from bcrypto package (https://github.com/bcoin-org/bcrypto) licensed under the MIT License.

final ZERO = Uint8List.fromList([0]);

class Signature {
  Uint8List _r;
  Uint8List _s;
  int _size;

  Signature(Uint8List r, Uint8List s) {
    this._r = r;
    this._s = s;
  }

  Signature.decode(Uint8List data, [int size = 32]) {
    this._size = size;
    if (data.length != this._size * 2) throw ArgumentError('Invalid data length');

    this.setR(data.sublist(0, this._size));
    this.setS(data.sublist(this._size, this._size * 2));
  }

  Signature.fromDER(Uint8List data, [int size = 32]) {
    this._size = size;

    int len = 0;
    int pos = 0;
    int rlen = 0;
    int slen = 0;
    Uint8List r;
    Uint8List s;

    if (pos + 1 > data.length) throw new Exception();
    if (data[pos] != 0x30) throw new Exception();
    pos += 1;

    if (pos + 1 > data.length) throw new Exception();
    len = data[pos];
    pos += 1;

    if (len & 0x80 != 0) {
      len -= 0x80;
      if (pos + len > data.length) throw new Exception();
      pos += len;
    }

    if (pos + 1 > data.length) throw new Exception();
    if (data[pos] != 0x02) throw new Exception();
    pos += 1;

    if (pos + 1 > data.length) throw new Exception();
    len = data[pos];
    pos += 1;

    if (len & 0x80 != 0) {
      len -= 0x80;
      if (pos + len > data.length) throw new Exception();

      while (len > 0 && data[pos] == 0x00) {
        len -= 1;
        pos += 1;
      }

      if (len > 6) throw new Exception();

      while (len > 0) {
        rlen *= 0x100;
        rlen += data[pos];
        len -= 1;
        pos += 1;
      }
    } else {
      rlen = len;
    }

    // Ignore leading zeroes in R.
    if (pos + rlen > data.length) throw new Exception();

    while (rlen > 0 && data[pos] == 0x00) {
      rlen -= 1;
      pos += 1;
    }

    if (rlen > this._size) {
      r = new Uint8List(this._size);
    } else {
      r = data.sublist(pos, pos + rlen);
    }

    pos += rlen;

    // Integer tag byte for S.
    if (pos + 1 > data.length) throw new Exception();
    if (data[pos] != 0x02) throw new Exception();
    pos += 1;

    // Integer length for S.
    if (pos + 1 > data.length) throw new Exception();
    len = data[pos];
    pos += 1;

    if (len & 0x80 != 0) {
      len -= 0x80;
      if (pos + len > data.length) throw new Exception();

      while (len > 0 && data[pos] == 0x00) {
        len -= 1;
        pos += 1;
      }

      if (len > 6) throw new Exception();

      while(len > 0) {
        slen *= 0x100;
        slen += data[pos];
        len -= 1;
        pos += 1;
      }
    } else {
      slen = len;
    }

    // Ignore leading zeroes in S.
    if (pos + slen > data.length) throw new Exception();

    while (slen > 0 && data[pos] == 0x00) {
      slen -= 1;
      pos += 1;
    }

    if (slen > this._size) {
      s = new Uint8List(this._size);
    } else {
      s = data.sublist(pos, pos + slen);
    }

    pos += slen;

    this.setR(r);
    this.setS(s);
  }

  Uint8List get r {
    return this._r;
  }

  Uint8List get s {
    return this._s;
  }

  setR(Uint8List r) {
    this._r = r;
  }

  setS(Uint8List s) {
    this._s = s;
  }

  Uint8List encode() {
    if (this.r.length != this._size) throw new ArgumentError('R length is wrong');
    if (this.s.length != this._size) throw new ArgumentError('S length is wrong');

    Buffer data = new Buffer();
    data.put(this.r);
    data.put(this.s);
    return data.toBytes();
  }

  Uint8List toDER() {
    Uint8List r = this._encodeInt(this.r);
    Uint8List s = this._encodeInt(this.s);

    if (r[0] & 0x80 != 0) throw new ArgumentError('R value is negative');
    if (s[0] & 0x80 != 0) throw new ArgumentError('S value is negative');
    if (r.length> 1 && (r[0] == 0x00) && r[1] & 0x80 == 0) throw new ArgumentError('R value excessively padded');
    if (s.length > 1 && (s[0] == 0x00) && s[1] & 0x80 == 0) throw new ArgumentError('S value excessively padded');
    
    int seq = 2 + r.length + 2 + s.length;
    int wide = seq >= 0x80 ? 1 : 0;
    int len = 2 + wide + seq;
    
    Buffer signature = new Buffer();

    signature.putInt(0x30);
    if (wide == 1) signature.putInt(0x81);
    signature.putInt(seq);
    signature.putInt(0x02);
    signature.putInt(r.length);
    signature.put(r);
    signature.putInt(0x02);
    signature.putInt(s.length);
    signature.put(s);

    Uint8List bytes = signature.toBytes();
    if (bytes.length != len) throw new Exception('Invalid signature length');

    return bytes;
  }

  Uint8List _encodeInt (Uint8List buf) {
    int i = 0;
    while (buf[i] == 0) ++i;
    if (i == buf.length) return ZERO;
    buf = buf.sublist(i);
    List<int> combine = List.from(ZERO);
    combine.addAll(buf);
    if (buf[0] & 0x80 != 0) return Uint8List.fromList(combine);
    return buf;
  }
}